/*
 * NCURSUI.C -- NCURSES USER INTERFACE UTILS
 * 
 * Copyright (C) 1998 by Ralph F. Siemsen
 * Released under the GNU public license
 *
 * Modified for Zaphod by Mark Knox, Feb 1999
 * 
 * Identification:
 *   $Id: ncursui.C,v 1.10 1999/04/22 17:55:11 jasona Exp $
 */

#include "ncursui.h"
#include <config.h>

extern char Stored[][FIELD_MAX];
extern char *Labels[];

/*** Window support routines ***/

int Window::window_count = 0;

Window::Window (int Left, int Top, int Width, int Height) {
  /* Initialize ncurses on the first window */
  if (window_count++ == 0) {
    initscr();
    cbreak();
    noecho();
    keypad (stdscr, TRUE);
    meta (stdscr, TRUE);
    timeout (100);
  }
  /* Adjust width and height dimensions if they are negative */
  while (Width <= 0) Width += COLS;
  while (Height <= 0) Height += LINES;
  
  /* Create the window */
  if (Width <= COLS && Height <= LINES) {
    left = Left; top = Top;
    width = Width; height = Height;
    win = newwin (Height, Width, Top, Left);
    if (win == 0) {
      endwin(); printf("window failed to open.\n"); exit (1);
    }
    keypad (win, TRUE);
    meta (win, TRUE);
    pan = new_panel (win);
    if (pan == 0) {
      endwin(); printf("panel failed to open\n"); exit(1);
    }
    wtimeout (win, 100);
  } else {
    endwin(); printf("Must be run in a 80x25 environment\n"); exit(1);
  }  
}

Window::~Window () {
  del_panel(pan);
  delwin(win);
  Update();
  if (--window_count == 0) endwin();
}

void Window::Box (int Draw, int Left, int Top, int Width, int Height) {
  /* Draw or erase a border in the window */
  int i;
  int ux = Left, uy = Top;
  int lx = Left + Width - 1, ly = Top + Height - 1;

  for (i = ux + 1; i < lx; i++)
    {
      mvwaddch (win, uy, i, Draw ? ACS_HLINE : 32);
      mvwaddch (win, ly, i, Draw ? ACS_HLINE : 32);
    }
  for (i = uy + 1; i < ly; i++)
    {
      mvwaddch (win, i, ux, Draw ? ACS_VLINE : 32);
      mvwaddch (win, i, lx, Draw ? ACS_VLINE : 32);
    }
  mvwaddch (win, uy, ux, Draw ? ACS_ULCORNER : 32);
  mvwaddch (win, uy, lx, Draw ? ACS_URCORNER : 32);
  mvwaddch (win, ly, ux, Draw ? ACS_LLCORNER : 32);
  mvwaddch (win, ly, lx, Draw ? ACS_LRCORNER : 32);
  wmove (win, 1, 1);
}

inline void Window::Update () { 
  update_panels(); 
  doupdate();
}

void Window::Border (int Draw) { 
  Box (Draw, 0, 0, width, height); 
}

void Window::Print (int Col, int Line, char *Text) {
  mvwprintw (win, Line, Col, Text);
}

void Window::Title (char *Title) {
  title = Title;
  if (left == 0 && top == 0 && width == COLS && height == LINES) {
    wattrset (win, A_REVERSE);
    mvwprintw (win, 0, 0, " %-*s", COLS-1, Title);
    wattrset (win, A_NORMAL);
  } else {
    wattrset (win, A_BOLD);
    mvwprintw (win, 0, 2, Title);
    wattrset (win, A_NORMAL);
  }
}

void Window::Help (char *Msg) {
  /* Place "help" message at the bottom of the screen */
  wattrset (win, A_REVERSE);
  mvwprintw (win, height - 1, 0, " %-*s", COLS-1, Msg);
  wattrset (win, A_NORMAL);
  touchline (win, height - 1, 0);
}

/*** Field Handling ***/

#define STRLEN(x) ((int)strlen(x))

Field::Field (Window *Win, int Row, int Col, int Width, int Id) 
  : win(Win), row(Row), col(Col), width(Width), id(Id)
{
  cursor = 0;
  changed = 0;
  // Copy the given value into internal buffer
  strncpy (value, Stored[Id], FIELD_MAX);
  value[FIELD_MAX-1] = 0;
  // Set the default cursor position.
  if (STRLEN(value) < width)
    cursor = STRLEN(value);
  else
    cursor = width;
  blankAllowed = false;
}

void Field::Help (Window *HelpWin, int MsgNum) {
  extern char *HelpMsg[];
  werase (HelpWin->win);
  HelpWin->Print (0, 0, HelpMsg[id]);
  HelpWin->Update();
}

int Field::ShowError (int Left, int Top, int MsgNum)
{
  int ch;
  extern char *ErrMsg[];
  {
    Window ErrWin(Left, Top, COLS-2*Left, LINES-2*Top);
    ErrWin.Border (1);
    Window ErrData (ErrWin.left+2,ErrWin.top+2,ErrWin.width-4,ErrWin.height-4);
    ErrData.Print (0,0,ErrMsg[MsgNum]);
    wattrset (ErrData.win, A_REVERSE);
    ErrData.Print (ErrData.width/2-2,ErrData.height-1,"<OK>");
    wattrset (ErrData.win, A_NORMAL);
    ErrData.Update();
    
    while ((ch = wgetch(ErrData.win)) == ERR);
    if (ch == 27) ungetch (27);
  }
  Display (0);		// This sequence places the cursor back
  win->Update();	// where it belongs on the widget that failed.
  return ch;
}

/* Text fields */

TextField::TextField (
  Window *Win, int Row, int Col, int Width, int Id)
  : Field (Win, Row, Col, Width, Id) {
  scroll = 0;
}

void TextField::Display (int Highlight) {
  // Adjust scroll position if needed:
  while (cursor < scroll) scroll--;
  while (cursor > scroll+width) scroll++;
  // Truncate value to fit in display
  char buffer[width+1];
  strncpy (buffer, value + scroll, width);
  buffer[width] = 0;
  // Print label
  win->Print (col - STRLEN(Labels[id]) - 2, row, Labels[id]);
  win->Print (col - 2, row, ": ");
  wmove (win->win, col, row);
  // Clear the destination area, then print value
  for (int i=0; i<width; i++) waddch (win->win, 32);
  if (Highlight) wattrset (win->win, A_REVERSE);
  win->Print (col, row, buffer);
  if (Highlight) wattrset (win->win, A_NORMAL);
  wmove (win->win, row, col + cursor - scroll);
  touchline (win->win, row, 1);
}

int TextField::Edit (int Ch) {
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT) {
    if (cursor > 0) cursor--;
  } else if (Ch == KEY_RIGHT) {
    if (cursor < STRLEN(value)) cursor++;
  } else if (Ch == KEY_HOME) {
    cursor = 0;
  } else if (Ch == KEY_END) {
    cursor = STRLEN(value);
  } else if (isalnum (Ch) || Ch == '-' || Ch == '_' || Ch == '.') {
    if (!changed) {
      cursor = 0;
      value[0] = 0;
    }
    if (STRLEN(value) < FIELD_MAX-1) {
      for (int i=STRLEN(value); i>=cursor; i--)
        value[i+1] = value[i];
      value[cursor] = Ch;
      cursor++;
    }
  } else if (Ch == 8 || Ch == 127 || Ch == KEY_BACKSPACE) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    if (cursor > 0) {
      for (int i=cursor, j=STRLEN(value); i<=j; i++)
        value[i-1] = value[i];
      cursor--;
    }
  } else if (Ch == KEY_DC) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    for (int i=cursor, j=STRLEN(value); i<j; i++)
      value[i] = value[i+1];
  } else {
    return Ch;
  }
  changed = 1;
  Display (0);
  win->Update();
  return 0;
}

int TextField::Validate (void) {
  if (!changed) return 1;
  if (!blankAllowed && STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  if (value[0] == '.' || value[STRLEN(value)-1] == '.') {
    ShowError (10,7,1);
    return 0;
  }
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

/* Number fields: text fields allowing only numeric characters */
NumberField::NumberField(
  Window *Win, int Row, int Col, int Width, int Id) 
  : TextField (Win, Row, Col, Width, Id) {
}

int NumberField::Edit (int Ch) {
  // Just like a regular text field, but only allow numbers
  if (isprint(Ch) && !isdigit(Ch)) Ch = 0;
  return TextField::Edit (Ch);
}

int NumberField::Validate (void) {
  // Don't error check if nothing changed
  if (!changed) return 1;
  // Put any additional error checks here.
  if (STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  // (return 0 if there is a error and you need to abort)
  // If everything is alright, then we should do:
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

/* Pathname fields: text field that allows / and \ */

PathField::PathField (
  Window *Win, int Row, int Col, int Width, int Id)
  : Field (Win, Row, Col, Width, Id) {
  scroll = 0;
}

void PathField::Display (int Highlight) {
  // Adjust scroll position if needed:
  while (cursor < scroll) scroll--;
  while (cursor > scroll+width) scroll++;
  // Truncate value to fit in display
  char buffer[width+1];
  strncpy (buffer, value + scroll, width);
  buffer[width] = 0;
  // Print label
  win->Print (col - STRLEN(Labels[id]) - 2, row, Labels[id]);
  win->Print (col - 2, row, ": ");
  wmove (win->win, col, row);
  // Clear the destination area, then print value
  for (int i=0; i<width; i++) waddch (win->win, 32);
  if (Highlight) wattrset (win->win, A_REVERSE);
  win->Print (col, row, buffer);
  if (Highlight) wattrset (win->win, A_NORMAL);
  wmove (win->win, row, col + cursor - scroll);
  touchline (win->win, row, 1);
}

int PathField::Edit (int Ch) {
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT) {
    if (cursor > 0) cursor--;
  } else if (Ch == KEY_RIGHT) {
    if (cursor < STRLEN(value)) cursor++;
  } else if (Ch == KEY_HOME) {
    cursor = 0;
  } else if (Ch == KEY_END) {
    cursor = STRLEN(value);
  } else if (isalnum (Ch) || Ch == '-' || Ch == '_' || Ch == '.' 
  		|| Ch == '/' || Ch == '\\') {
    if (!changed) {
      cursor = 0;
      value[0] = 0;
    }
    if (STRLEN(value) < FIELD_MAX-1) {
      for (int i=STRLEN(value); i>=cursor; i--)
        value[i+1] = value[i];
      value[cursor] = Ch;
      cursor++;
    }
  } else if (Ch == 8 || Ch == 127 || Ch == KEY_BACKSPACE) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    if (cursor > 0) {
      for (int i=cursor, j=STRLEN(value); i<=j; i++)
        value[i-1] = value[i];
      cursor--;
    }
  } else if (Ch == KEY_DC) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    for (int i=cursor, j=STRLEN(value); i<j; i++)
      value[i] = value[i+1];
  } else {
    return Ch;
  }
  changed = 1;
  Display (0);
  win->Update();
  return 0;
}

int PathField::Validate (void) {
  if (!changed) return 1;
  if (!blankAllowed && STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  if (value[STRLEN(value)-1] != '/') {
    ShowError (10,7,7);
    return 0;
  }
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

/* NumberList field: text field that allows only 0-9 comma and space */

NumberListField::NumberListField (
  Window *Win, int Row, int Col, int Width, int Id)
  : Field (Win, Row, Col, Width, Id) {
  scroll = 0;
}

void NumberListField::Display (int Highlight) {
  // Adjust scroll position if needed:
  while (cursor < scroll) scroll--;
  while (cursor > scroll+width) scroll++;
  // Truncate value to fit in display
  char buffer[width+1];
  strncpy (buffer, value + scroll, width);
  buffer[width] = 0;
  // Print label
  win->Print (col - STRLEN(Labels[id]) - 2, row, Labels[id]);
  win->Print (col - 2, row, ": ");
  wmove (win->win, col, row);
  // Clear the destination area, then print value
  for (int i=0; i<width; i++) waddch (win->win, 32);
  if (Highlight) wattrset (win->win, A_REVERSE);
  win->Print (col, row, buffer);
  if (Highlight) wattrset (win->win, A_NORMAL);
  wmove (win->win, row, col + cursor - scroll);
  touchline (win->win, row, 1);
}

int NumberListField::Edit (int Ch) {
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT) {
    if (cursor > 0) cursor--;
  } else if (Ch == KEY_RIGHT) {
    if (cursor < STRLEN(value)) cursor++;
  } else if (Ch == KEY_HOME) {
    cursor = 0;
  } else if (Ch == KEY_END) {
    cursor = STRLEN(value);
  } else if (isdigit (Ch) || Ch == ' ' || Ch == ',') { 
    if (!changed) {
      cursor = 0;
      value[0] = 0;
    }
    if (STRLEN(value) < FIELD_MAX-1) {
      for (int i=STRLEN(value); i>=cursor; i--)
        value[i+1] = value[i];
      value[cursor] = Ch;
      cursor++;
    }
  } else if (Ch == 8 || Ch == 127 || Ch == KEY_BACKSPACE) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    if (cursor > 0) {
      for (int i=cursor, j=STRLEN(value); i<=j; i++)
        value[i-1] = value[i];
      cursor--;
    }
  } else if (Ch == KEY_DC) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    for (int i=cursor, j=STRLEN(value); i<j; i++)
      value[i] = value[i+1];
  } else {
    return Ch;
  }
  changed = 1;
  Display (0);
  win->Update();
  return 0;
}

int NumberListField::Validate (void) {
  if (!changed) return 1;
  if (STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  if (value[0] == '.' || value[STRLEN(value)-1] == '.') {
    ShowError (10,7,1);
    return 0;
  }
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}


HostNameField::HostNameField (
  Window *Win, int Row, int Col, int Width, int Id) 
  : TextField (Win, Row, Col, Width, Id) {
}

int HostNameField::Edit (int Ch) {
  // Just like a regular text field, but don't allow '.' character
  if (Ch == '.') Ch = 0;
  return TextField::Edit (Ch);
}

int HostNameField::Validate (void) {
  // Don't error check if nothing changed
  if (!changed) return 1;
  // Put any additional error checks here.
  if (STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  // (return 0 if there is a error and you need to abort)
  // If everything is alright, then we should do:
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

EmailField::EmailField (
  Window *Win, int Row, int Col, int Width, int Id) 
  : TextField (Win, Row, Col, Width, Id) {
}

int EmailField::Edit (int Ch)
{
	// don't allow '*' character
	if (Ch == '*') Ch = 0;

	// Allow '@' char
	if (Ch == '@')
	{
		if (!changed)
		{
			cursor = 0;
			value[0] = 0;
		}
		if (STRLEN(value) < FIELD_MAX-1)
		{
			for (int i=STRLEN(value); i>=cursor; i--)
				value[i+1] = value[i];
			value[cursor] = Ch;
			cursor++;
		}
		changed = 1;
		Display (0);
		win->Update();
		return Ch;
	}
	else
		return TextField::Edit(Ch);
}

int EmailField::Validate (void) {
  // Don't error check if nothing changed
  if (!changed) return 1;
  // Put any additional error checks here.
  if (!blankAllowed && STRLEN(value) == 0) {
    ShowError (10,7,0);
    return 0;
  }
  // the email address must contain an '@'
  if (strchr(value, '@') == NULL)
  {
    ShowError (10,7,6);
    return 0;
  }
  // (return 0 if there is a error and you need to abort)
  // If everything is alright, then we should do:
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

IPField::IPField (
  Window *Win, int Row, int Col, int Width, int Id)
  : TextField (Win, Row, Col, Width, Id) {
}

int IPField::Edit (int Ch) {
  // Like a regular text field, but only numbers and periods
  if (Ch >= 32 && Ch < '0' && Ch != '.') return Ch;
  if (Ch > '9' && Ch < 127) return Ch;
  return TextField::Edit (Ch);
}

int IPField::Validate (void)
{
	// Don't error check if nothing changed
	if (!changed)
		return 1;
	
	// Handle special cases
	if ((id >= EXTERNAL_IPADDR) && (id <= INTERNAL_NETMASK))
	{
		if (strcmp (value, "0") == 0)
			strcpy (value, "(disabled)");
		if (strcmp (value, "(disabled)") == 0)
		{
			changed = 0;
			strcpy (Stored[id], value);
			cursor = STRLEN (Stored[id]);
      
			// Also reset the other pair member:
			int partner[] = { 3, 2, 5, 4 };
			strcpy (Stored[partner[id-EXTERNAL_IPADDR]-2+EXTERNAL_IPADDR],
				"(disabled)");
			extern Form *Sec;
			Field *part = Sec->items[partner[id-EXTERNAL_IPADDR]];
			strcpy (part->value, "(disabled)");
			part->cursor = STRLEN (part->value);
			return 1;
		}
	}

	// Ensure that it's a valid IP address
	int a,b,c,d;
	if (sscanf (value, "%d.%d.%d.%d", &a, &b, &c, &d) != 4)
	{
		ShowError (10,7,2);
		return 0;
	}
  
	if (	a<0 || a>255 || b<0 || b>255 ||
      		c<0 || c>255 || d<0 || d>255)
	{
		ShowError (10,7,2);
		return 0;
	}
  
	if (STRLEN(value) == 0)
	{
		ShowError (10,7,0);
		return 0;
	}

	// (return 0 if there is a error and you need to abort)
	// If everything is alright, then we should do:
	changed = 0;
	strcpy (Stored[id], value);
	return 1;
}

MultiIPField::MultiIPField (
  Window *Win, int Row, int Col, int Width, int Id)
  : TextField (Win, Row, Col, Width, Id) {
}

int MultiIPField::Edit (int Ch) {
  // This is exactly the same as TextField::Edit except that
  // only numbers/dots/spaces are allowed:
  
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT) {
    if (cursor > 0) cursor--;
  } else if (Ch == KEY_RIGHT) {
    if (cursor < STRLEN(value)) cursor++;
  } else if (Ch == KEY_HOME) {
    cursor = 0;
  } else if (Ch == KEY_END) {
    cursor = STRLEN(value);
  } else if (isdigit (Ch) || Ch == '.' || Ch == ' ') {
    if (!changed) {
      cursor = 0;
      value[0] = 0;
    }
    if (STRLEN(value) < FIELD_MAX-1) {
      for (int i=STRLEN(value); i>=cursor; i--)
        value[i+1] = value[i];
      value[cursor] = Ch;
      cursor++;
    }
  } else if (Ch == 8 || Ch == 127 || Ch == KEY_BACKSPACE) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    if (cursor > 0) {
      for (int i=cursor, j=STRLEN(value); i<=j; i++)
        value[i-1] = value[i];
      cursor--;
    }
  } else if (Ch == KEY_DC) {
    //if (!changed) {
    //  value[0] = 0;
    //  cursor = 0;
    //}
    for (int i=cursor, j=STRLEN(value); i<j; i++)
      value[i] = value[i+1];
  } else {
    return Ch;
  }
  changed = 1;
  Display (0);
  win->Update();
  return 0;
}

int MultiIPField::Validate (void) {
  // Don't error check if nothing changed
  if (!changed) return 1;
  // There could be multiple IP addresses:
  int ipcount = 0;
  int ipaddr[FIELD_MAX/9];
  char *ptr = value;
  while (*ptr != 0) {
    // Ensure that it's a valid IP address
    int a,b,c,d;
    if (sscanf (ptr, "%d.%d.%d.%d", &a, &b, &c, &d) != 4) {
      ShowError (10,7,2);
      return 0;
    }
    if (a<0 || a>255 || b<0 || b>255 ||
        c<0 || c>255 || d<0 || d>255) {
      ShowError (10,7,2);
      return 0;
    }
    if (STRLEN(value) == 0) {
      ShowError (10,7,0);
      return 0;
    }
    // Store it into table
    ipaddr[ipcount++] = (a << 24) | (b << 16) | (c << 8) | d;
    // Advance pointer to next item in string
    while ((*ptr == '.') || ((*ptr >= '0') && (*ptr <= '9'))) ptr++;
    while (*ptr == ' ') ptr++;
  }
  // Now re-write the "validated" string
  extern char *itod (int i);
//  strcpy (value, itod (ipaddr[0]));
  for (int i=1; i<ipcount; i++) {
    strcat (value, " ");
//    strcat (value, itod (ipaddr[i]));
  }
  // Return sucess to the caller routines
  changed = 0;
  strcpy (Stored[id], value);
  return 1;
}

/* List selection widget */

SelectField::SelectField (Window *Win, int Row, int Col, 
  int Width, char **Options, int Id)
  : Field (Win, Row, Col, Width, Id) {
  options = Options;
  for (count=0; options[count]; count++);
  choice = 0;		// default value
  // scan list and try to match specified value
  for (int i=0; i<count; i++) {
    if (strcmp(options[i], Stored[Id]) == 0) {
      choice = i;
      break;
    }
  }
}

void SelectField::Display (int Highlight) {
  // Truncate value to fit in display
  char buffer[width+1];
  strncpy (buffer, options[choice], width);
  buffer[width] = 0;
  // Print label
  win->Print (col - STRLEN(Labels[id]) - 2, row, Labels[id]);
  win->Print (col - 2, row, ": ");
  wmove (win->win, col, row);
  // Clear the destination area, then print value
  for (int i=0; i<width; i++) waddch (win->win, 32);
  if (Highlight) wattrset (win->win, A_REVERSE);
  win->Print (col, row, buffer);
  if (Highlight) wattrset (win->win, A_NORMAL);
  wmove (win->win, row, col + cursor);
  touchline (win->win, row, 1);
}

int SelectField::Edit (int Ch) {
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT || Ch == '-' || Ch == '_') {
    choice--;
    if (choice < 0) choice = count-1;
    cursor = STRLEN (options[choice]);
  } else if (Ch == KEY_RIGHT || Ch == '+' || Ch == '=') {
    choice++;
    if (choice >= count) choice = 0;
    cursor = STRLEN (options[choice]);
  } else if (isalnum (Ch)) {
    // try to match field
  } else {
    return Ch;
  }
  changed = 1;
  Display (0);
  win->Update();
  return 0;
}

int SelectField::Validate (void) {
  if (!changed) return 1;
  changed = 0;
  strcpy (Stored[id], options[choice]);
  return 1;
}

/* Special selection widget - F1 hotkey */

SpecialSelectField::SpecialSelectField (Window *Win, int Row, int Col, 
  int Width, fptr Action, char **Options, int Id)
  : SelectField (Win, Row, Col, Width, Options, Id) {
  action = Action;
}

int SpecialSelectField::Edit (int Ch) {
  // Deal with special keys
  if (Ch == KEY_F(1) || Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d) {
    Display(0);
    action();
    changed = 0;
    return 0;
  }
  return SelectField::Edit (Ch);
}

int SpecialSelectField::Validate (void) {
  if (!changed) {
    changed = 0;
    strcpy (Stored[id], options[choice]);
    return 1;
  } else {
    ungetch (KEY_F(1));		// make submenu appear
    changed = 0;
    strcpy (Stored[id], options[choice]);
    return 0;
  } 
}

/* Time entry widget */

TimeField::TimeField (Window *Win, int Row, int Col, int Id)
  : Field (Win, Row, Col, 20, Id) {
  cursor = 0;
  offset = 0;
}

void TimeField::Display (int Highlight) {
  // Print label
  win->Print (col - STRLEN(Labels[id]) - 2, row, Labels[id]);
  win->Print (col - 2, row, ": ");
  extern char *MonthOptions[];
  if (!Highlight) {
    // Field is not active - just display current time
    now = time(NULL) + offset;
    tm = *localtime(&now);
    sprintf (value, "%4d %s %02d %02d:%02d:%02d", tm.tm_year + 1900,
      MonthOptions[tm.tm_mon], tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec);
    win->Print (col, row, value);
    touchline (win->win, row, 1);
    win->Update ();
  } else {
    // Field has input focus - highlight current "item"
    wmove (win->win, col, row);
    if (cursor == 0) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%4d", tm.tm_year + 1900);
    if (cursor == 0) wattrset (win->win, A_NORMAL);
    waddch (win->win, 32);
    if (cursor == 1) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%s", MonthOptions[tm.tm_mon]);
    if (cursor == 1) wattrset (win->win, A_NORMAL);
    waddch (win->win, 32);
    if (cursor == 2) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%02d", tm.tm_mday);
    if (cursor == 2) wattrset (win->win, A_NORMAL);
    waddch (win->win, 32);
    if (cursor == 3) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%02d", tm.tm_hour);
    if (cursor == 3) wattrset (win->win, A_NORMAL);
    waddch (win->win, ':');
    if (cursor == 4) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%02d", tm.tm_min);
    if (cursor == 4) wattrset (win->win, A_NORMAL);
    waddch (win->win, ':');
    if (cursor == 5) wattrset (win->win, A_REVERSE);
    wprintw (win->win, "%02d", tm.tm_sec);
    if (cursor == 5) wattrset (win->win, A_NORMAL);
    const int cpos[6] = { 0, 5, 9, 12, 15, 18 };
    wmove (win->win, row, col + cpos[cursor]);
    touchline (win->win, row, 1);
  }
}

int TimeField::Edit (int Ch) {
  // Keycode "corrections"
  if (changed && (Ch == KEY_ENTER || Ch == 0x0a || Ch == 0x0d))
    Ch = KEY_DOWN;
  // Being keycode processing
  if (Ch == KEY_LEFT) {
    if (cursor > 0) cursor--;
  } else if (Ch == KEY_RIGHT) {
    if (cursor < 5) cursor++;
  } else if (Ch == '+' || Ch == '=') {
    Adjust (cursor, +1);
  } else if (Ch == '-' || Ch == '_') {
    Adjust (cursor, -1);
  } else if (isalnum (Ch)) {
    // try to match field
  } else {
    return Ch;
  }
  changed = 1;
  Display (1);
  win->Update();
  return 0;
}

void TimeField::Adjust (int Field, int Amount) {
  switch (Field) {
    case 0:
      tm.tm_year += Amount;
      break;
    case 1:
      tm.tm_mon += Amount;
      while (tm.tm_mon < 0) tm.tm_mon += 12;
      while (tm.tm_mon > 11) tm.tm_mon -= 12;
      break;
    case 2:
      tm.tm_mday += Amount;
      while (tm.tm_mday-1 < 0) tm.tm_mday += 31;
      while (tm.tm_mday-1 > 30) tm.tm_mday -= 31;
      break;
    case 3:
      tm.tm_hour += Amount;
      while (tm.tm_hour < 0) tm.tm_hour += 24;
      while (tm.tm_hour > 23) tm.tm_hour -= 24;
      break;
    case 4:
      tm.tm_min += Amount;
      while (tm.tm_min < 0) tm.tm_min += 60;
      while (tm.tm_min > 59) tm.tm_min -= 60;
      break;
    case 5:
      tm.tm_sec += Amount;
      while (tm.tm_sec < 0) tm.tm_sec += 60;
      while (tm.tm_sec > 59) tm.tm_sec -= 60;
      break;
  }
}

int TimeField::Validate (void) {
  // We could do some more sanity checks here
  if (tm.tm_year < 1 || tm.tm_year > 137) {
    ShowError (10,7,5);
    return 0;
  }
  time_t newtime = mktime (&tm);
  if (newtime == (time_t) -1) {
    ShowError (10,7,4);
    return 0;
  }
  // Compute delta from system time
  now = time(NULL);
  offset = newtime - now;
  changed = 0;
  return 1;
}

/*** Form Handling ***/

Form::Form (Field **Forms, Window *HelpWin)
  : items(Forms), helpwin(HelpWin) {
  for (count=0; items[count]; count++);
  selected = 0;
}

Form::~Form (void) {
}

int Form::Edit (int Ch) {
  // Let the child widget have the input first
  Ch = items[selected]->Edit (Ch);
  // Then deal with "standard" cases
  switch (Ch) {
    case KEY_LEFT:
    case KEY_UP:
    case KEY_BTAB:
      if (selected > 0) {
        if (items[selected]->Validate()) {
          selected--;
	  Render();
	}
      }
      break;
    case KEY_RIGHT:
    case KEY_DOWN:
    case 0x09:
      if (selected < count-1) {
        if (items[selected]->Validate()) {
          selected++;
          Render();
        }
      }
      break;
    case KEY_F(10):
      if (items[selected]->Validate())
        return Ch;
      break;
  }
  return 0;
}

void Form::Render (void) {
  for (int i=0; i<count; i++)
    items[i]->Display(0);
  items[selected]->Help (helpwin, selected);
  items[selected]->Display(1);
  items[selected]->win->Update();
}

/*** File support routines ***/

File::File (char *name) {
  fh = fopen (name, "r");
}

File::~File () {
  if (fh) fclose (fh);
}

char *File::ReadParam (char *name) {
  static char result[FIELD_MAX];
  result[0] = 0;
  if (fh) {
    char buf[FIELD_MAX];
    rewind (fh);
    while (fgets(buf,FIELD_MAX,fh) != 0) {
      // Look for key string, followed by '=' symbol
      if ((strncmp(buf,name,strlen(name)) == 0) &&
          ((buf[strlen(name)] == '=') || 
           (buf[strlen(name)] == 32) ||
           (buf[strlen(name)] == 9))) {
        // Find beginning of string to copy
        int i = strlen(name);
        while ((buf[i] == '=') || (buf[i] == ' ') || (buf[i] == '\t')) i++;
        // Now copy characters until end of token
        int j = strlen(result);
        if (j > 0) result[j++] = 32;
        while ((buf[i] != ' ') && (buf[i] != '\t') &&
 	  (buf[i] != 10) && (buf[i] != 13))
        result[j++] = buf[i++];
        // Null terminate the string
        result[j] = 0;
      }
    }
    return result;
  }
  // If we get here, the function has failed.
  if ((strcmp(name,"IPADDR") == 0) || (strcmp(name,"NETMASK") == 0))
    return "(disabled)";
  else if (strcmp(name,"HOSTNAME") == 0)
    return "yourhostname";
  else if (strcmp(name,"DOMAINNAME") == 0)
    return "yourdomain.com";
  else return "";
}

FWrite::FWrite (char *name) {
  char backup[1024];
  sprintf (backup, "mv %s %s.old >/dev/null 2>&1", name, name);
  system (backup);
  fh = fopen (name, "w");
}

FWrite::~FWrite () {
  if (fh) fclose (fh);
}

void FWrite::Printf (char *msg)
{
  if (fh) fprintf (fh, msg);
}

/*** Menu support ***/

/* MENU SELECTION ROUTINE */

int Window::DrawMenu (Menu *menu)
{
  int y, x;
  int col, line;
  int length, maxlength = 0;
  int n, left;

  /* Determine length of longest item in the menu */
  getyx (win, y, x);
	y = 1;
  for (n = 0; n < menu->Lines * menu->Cols; n++)
    {
      length = 0;
      if (menu->Item[n]->Prefix)
	length += strlen (menu->Item[n]->Prefix);
      if (menu->Item[n]->Hotkey)
	length += 1;
      if (menu->Item[n]->Suffix)
	length += strlen (menu->Item[n]->Suffix);
      if (length > maxlength)
	maxlength = length;
    }

  /* Now go and print the menu out */
  maxlength += 4;		/* Leave room around sides */
  getmaxyx (win, n, left);
  left = (left - menu->Cols * (maxlength + menu->Gap)) / 2;
  left += (maxlength + menu->Gap) / 2;

  for (col = 0; col < menu->Cols; col++)
    for (line = 0; line < menu->Lines; line++)
      {
	n = col * menu->Lines + line;

	/* Recalculate the length of this item, for centering */
	length = 0;
	if (menu->Item[n]->Prefix)
	  length += strlen (menu->Item[n]->Prefix);
	if (menu->Item[n]->Hotkey)
	  length += 1;
	if (menu->Item[n]->Suffix)
	  length += strlen (menu->Item[n]->Suffix);

	/* Move cursor and draw the item */
	wmove (win, y + line * 2, left + (maxlength + menu->Gap) * col - length / 2);
	if (menu->Item[n]->Prefix)
	  wprintw (win, menu->Item[n]->Prefix);
	wattrset (win, A_UNDERLINE);
	if (menu->Item[n]->Hotkey)
	  wprintw (win, "%c", menu->Item[n]->Hotkey);
	wattrset (win, A_NORMAL);
	if (menu->Item[n]->Suffix)
	  wprintw (win, menu->Item[n]->Suffix);
      }

  wmove (win, y - 1, left);
  return (maxlength);
}

int Window::Select (Menu *menu, int pre)
{
  int col = pre / menu->Lines;
  int line = pre % menu->Lines;
  int y, x;
  int c, n;
  int left, boxlen, colwidth;

  /* Draw menu and establish box-drawing positions */
  boxlen = DrawMenu (menu);
  colwidth = boxlen + menu->Gap;
  getmaxyx (win, n, left);
  left = (left - menu->Cols * colwidth) / 2 + menu->Gap / 2;

  getyx (win, y, x);
  y = 1;
  for (;;)
    {
      Box (TRUE, left + col * colwidth, y - 1+ line  * 2, boxlen, 3);
      wmove (win, y + line * 2, left + col * colwidth + boxlen - 2);
      c = wgetch (win);
      Box (FALSE, left + col * colwidth, y - 1+ line  * 2, boxlen, 3);
      switch (c)
	{
	case KEY_UP:
	  if (line > 0)
	    line--;
	  break;
	case KEY_DOWN:
	  if (line < menu->Lines - 1)
	    line++;
	  break;
	case KEY_LEFT:
	  if (col > 0)
	    col--;
	  break;
	case KEY_RIGHT:
	  if (col < menu->Cols - 1)
	    col++;
	  break;
	case 10:
	case 13:
	case KEY_ENTER:
	  return (line + col * menu->Lines);
	  break;
	default:
	  for (n = 0; n < menu->Lines * menu->Cols; n++)
	    if (toupper (menu->Item[n]->Hotkey) == toupper (c))
	      return n;
//	  beep ();
	  break;
	}
    }
}

